/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2010                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
 |      version: 0.1 04-01-2010                                             |
 |                                                                          |
 \*--------------------------------------------------------------------------*/

#include "mm.hh"
#include "timing.hh"
//#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <cmath>
#include <limits>

using namespace std ;

#define A(I,J) ADDR(Amat,I,J)
#define B(I,J) ADDR(Bmat,I,J)

static
void
M_check( char const msg[],
         valueType * Amat, indexType ldAmat, 
         valueType * Bmat, indexType ldBmat,
         indexType n,
         indexType m ) {
  valueType res = 0 ;
  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType j = 0 ; j < m ; ++j )
      res = std::max( res, abs( A(i,j)-B(i,j) ) ) ;
  
  if ( res > numeric_limits<valueType>::epsilon()*1000 ) {
    cout << "\n\n\nerror in matrix multiplication: " << msg << " = " << res << "\n\n\n" ;
    exit(0) ;
  }
}

int
main() {

  Timing tm ;

  valueType *A, *B, *C, *Cref, elapsed ;

  A    = new valueType[ 5000 * 5000 ] ;
  B    = new valueType[ 5000 * 5000 ] ;
  C    = new valueType[ 5000 * 5000 ] ;
  Cref = new valueType[ 5000 * 5000 ] ;

  cout << setw(6)  << "Size"       ;
  cout << setw(12) << "normal"     ; // 1
  cout << setw(12) << "temp"       ; // 2
  #if 0
  cout << setw(12) << "unroll4"    ; // 3
  cout << setw(12) << "unroll8"    ; // 4
  cout << setw(12) << "unroll16"   ; // 5
  #endif
  cout << setw(12) << "pointer"    ; // 6
  cout << setw(12) << "lam"        ; // 7
  cout << setw(12) << "tiling4"    ; // 8
  cout << setw(12) << "tiling8"    ; // 9
  cout << setw(12) << "tiling16"   ; // 10
  #if 0
  cout << setw(12) << "Maeno4"     ; // 11
  cout << setw(12) << "Maeno8"     ; // 12
  cout << setw(12) << "Maeno16"    ; // 13
  #endif
  cout << setw(12) << "Warner4"    ; // 14
  cout << setw(12) << "Warner8"    ; // 15
  cout << setw(12) << "Warner16"   ; // 16
  cout << setw(12) << "Blas"       ; // 17
  #if 0
  cout << setw(12) << "blocking16" ; // 18
  #endif
  cout << setw(12) << "strassen"   ; // 19
  cout << '\n' ;

  for ( indexType N = 100 ; N <= 600 ; N += 100 ) {

    cout << setw(6) << N ;
    for ( indexType i = 0 ; i < N ; ++i )
      for ( indexType j = 0 ; j < N ; ++j )
        A[i*N+j] = (arc4random()%10000) / 22123.0 ;
    for ( indexType i = 0 ; i < N ; ++i )
      for ( indexType j = 0 ; j < N ; ++j )
        B[i*N+j] = (arc4random()%11111) / 32455.0 ;
    //cout << "Generated matrix\n" ;

    tm . start() ;
    MM_normal( A, N, B, N, Cref, N, N, N, N ) ; // 1
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;

    tm . start() ;
    MM_temp( A, N, B, N, C, N, N, N, N ) ; // 2
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "temp", C, N, Cref, N, N, N ) ;

    #if 0
    tm . start() ;
    MM_unroll4( A, N, B, N, C, N, N, N, N ) ; // 3
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "unroll4", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_unroll8( A, N, B, N, C, N, N, N, N ) ; // 4
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "unroll8", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_unroll16( A, N, B, N, C, N, N, N, N ) ; // 5
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "unroll16", C, N, Cref, N, N, N ) ;
    #endif

    tm . start() ;
    MM_pointer( A, N, B, N, C, N, N, N, N ) ; // 6
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "pointer", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_lam( A, N, B, N, C, N, N, N, N ) ; // 7
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "lam", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_tiling( A, N, B, N, C, N, N, N, N, 4 ) ; // 8
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "tiling4", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_tiling( A, N, B, N, C, N, N, N, N, 8 ) ;
    elapsed = tm . milliseconds() ; // 9
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "tiling8", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_tiling( A, N, B, N, C, N, N, N, N, 16 ) ; // 10
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "tiling16", C, N, Cref, N, N, N ) ;

    #if 0
    tm . start() ;
    MM_maeno( A, N, B, N, C, N, N, N, N, 4 ) ; // 11
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "maeno4", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_maeno( A, N, B, N, C, N, N, N, N, 8 ) ; // 12
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "maeno8", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_maeno( A, N, B, N, C, N, N, N, N, 16 ) ; // 13
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "maeno16", C, N, Cref, N, N, N ) ;
    #endif

    tm . start() ;
    MM_warner( A, N, B, N, C, N, N, N, N, 4 ) ; // 14
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "warner4", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_warner( A, N, B, N, C, N, N, N, N, 8 ) ; // 15
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "warner8", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_warner( A, N, B, N, C, N, N, N, N, 16 ) ; // 16
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "warner16", C, N, Cref, N, N, N ) ;

    tm . start() ;
    MM_blas( A, N, B, N, C, N, N, N, N ) ; // 17
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "blas", C, N, Cref, N, N, N ) ;

    #if 0
    tm . start() ;
    MM_blocking( A, N, B, N, C, N, N, N, N, 16 ) ; // 18
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "blocking16", C, N, Cref, N, N, N ) ;
    #endif

    tm . start() ;
    MM_strassen( A, N, B, N, C, N, N, N, N ) ; // 19
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "strassen", C, N, Cref, N, N, N ) ;

    cout << '\n' ;
  
  }
  
  delete [] A ;
  delete [] B ;
  delete [] C ;
  delete [] Cref ;

  return 0 ;
}

// EOF main.cc
