/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2010                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
 |      version: 0.1 04-01-2010                                             |
 |                                                                          |
 \*--------------------------------------------------------------------------*/

#include "mm.hh"
#include "timing.hh"
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <cmath>

using namespace std ;

#define A(I,J) ADDR(Amat,I,J)
#define B(I,J) ADDR(Bmat,I,J)

void
M_check( char const msg[],
         valueType * Amat, indexType ldAmat, 
         valueType * Bmat, indexType ldBmat,
         indexType n,
         indexType m ) {
  valueType res = 0 ;
  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType j = 0 ; j < m ; ++j )
      res = std::max( res, abs( A(i,j)-B(i,j) ) ) ;
  
  if ( res > 1E-6 ) {
    cout << "\n\n\nerror in matrix multiplication: " << msg << " = " << res << "\n\n\n" ;
    exit(0) ;
  }
}

int
main() {

  Timing tm ;

  valueType *A, *B, *C, *Cref, elapsed ;

  A    = new valueType[ 6000 * 6000 ] ;
  B    = new valueType[ 6000 * 6000 ] ;
  C    = new valueType[ 6000 * 6000 ] ;
  Cref = new valueType[ 6000 * 6000 ] ;

  for ( indexType i = 0 ; i < 6000 * 6000 ; ++i ) A[i] = arc4random() * 1.02323E-8 ;
  for ( indexType i = 0 ; i < 6000 * 6000 ; ++i ) B[i] = arc4random() * 2.323E-8 ;

  cout << setw(6)  << "Size"       ;
  cout << setw(12) << "Blas"       ;
  cout << setw(12) << "strassen"   ;
  cout << '\n' ;

  for ( indexType N = 6000 ; N >= 1000 ; N *= 0.5 ) {

    cout << setw(6) << N ;

    indexType m = N-1 ;
    indexType p = N-1 ;
    indexType n = N-3 ;

    tm . start() ;
    MM_blas( A, N, B, N, Cref, N, m, p, n ) ;
    //MM_strassen( A, N, B, N, Cref, N, m, p, n ) ;
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;

    tm . start() ;
    MM_strassen( A, N, B, N, C, N, m, p, n ) ;
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed  << "ms" ;
    M_check( "strassen", C, N, Cref, N, m, n ) ;

    cout << '\n' ;
  
  }
  
  delete [] A ;
  delete [] B ;
  delete [] C ;
  delete [] Cref ;

  return 0 ;
}

// EOF main.cc
