/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2010                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
 |      version: 0.1 04-01-2010                                             |
 |                                                                          |
\*--------------------------------------------------------------------------*/

#include "mm.hh"
#include "alglin.hh"
#include "strassen.hh"
#include <algorithm>

using namespace std ;

/*
 * matrix multiply tests -- C language, version 1.0, May 1993
 *
 * Contact Mark Smotherman (mark@cs.clemson.edu)
 */

#define A(I,J) ADDR(a,I,J)
#define B(I,J) ADDR(b,I,J)
#define C(I,J) ADDR(c,I,J)

///////////////////////////////////////////////////////

void
MM_normal( valueType const a[], indexType lda, // n x p
           valueType const b[], indexType ldb, // p x m
           valueType       c[], indexType ldc,
           indexType n,
           indexType p,
           indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i ) {
    for ( indexType j = 0 ; j < m ; ++j ) {
	    C(i,j) = 0 ;
	    for ( indexType k = 0 ; k < p ; ++k ) {
        C(i,j) += A(i,k)*B(k,j) ;
      }
    }
  }
}

///////////////////////////////////////////////////////

void
MM_temp( valueType const a[], indexType lda, // n x p
         valueType const b[], indexType ldb, // p x m
         valueType       c[], indexType ldc,
         indexType n,
         indexType p,
         indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i ) {
    for ( indexType j = 0 ; j < m ; ++j ) {
	    valueType temp = 0 ;
	    for ( indexType k = 0 ; k < p ; ++k ) {
        temp += A(i,k)*B(k,j) ;
      }
	    C(i,j) = temp ;
    }
  }
}

///////////////////////////////////////////////////////

void
MM_unroll4( valueType const a[], indexType lda, // n x p
            valueType const b[], indexType ldb, // p x m
            valueType       c[], indexType ldc,
            indexType n,
            indexType p,
            indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i ) {
    for ( indexType j = 0 ; j < m ; ++j ) {
	    indexType k    = 0 ;
	    valueType temp = 0 ;
	    for ( ; k < (p-3) ; k += 4 ) {
        temp += A(i,k+0) * B(k+0,j) ;
        temp += A(i,k+1) * B(k+1,j) ;
        temp += A(i,k+2) * B(k+2,j) ;
        temp += A(i,k+3) * B(k+3,j) ;
      }
	    for ( ; k < p ; ++k ) temp += A(i,k)*B(k,j) ;
	    C(i,j) = temp ;
    }
  }
}

///////////////////////////////////////////////////////

void
MM_unroll8( valueType const a[], indexType lda, // n x p
            valueType const b[], indexType ldb, // p x m
            valueType       c[], indexType ldc,
            indexType n,
            indexType p,
            indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i ) {
    for ( indexType j = 0 ; j < m ; ++j ) {
	    indexType k    = 0 ;
	    valueType temp = 0 ;
	    for ( ; k < (p-7) ; k += 8 ) {
        temp += A(i,k+0) * B(k+0,j) ;
        temp += A(i,k+1) * B(k+1,j) ;
        temp += A(i,k+2) * B(k+2,j) ;
        temp += A(i,k+3) * B(k+3,j) ;
        temp += A(i,k+4) * B(k+4,j) ;
        temp += A(i,k+5) * B(k+5,j) ;
        temp += A(i,k+6) * B(k+6,j) ;
        temp += A(i,k+7) * B(k+7,j) ;
      }
	    for ( ; k < p ; ++k ) temp += A(i,k)*B(k,j) ;
	    C(i,j) = temp ;
    }
  }
}

///////////////////////////////////////////////////////

void
MM_unroll16( valueType const a[], indexType lda, // n x p
             valueType const b[], indexType ldb, // p x m
             valueType       c[], indexType ldc,
             indexType n,
             indexType p,
             indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i ) {
    for ( indexType j = 0 ; j < m ; ++j ) {
	    indexType k    = 0 ;
	    valueType temp = 0 ;
	    for ( ; k < (p-15) ; k += 16 ) {
        temp += A(i,k+ 0) * B(k+ 0,j) ;
        temp += A(i,k+ 1) * B(k+ 1,j) ;
        temp += A(i,k+ 2) * B(k+ 2,j) ;
        temp += A(i,k+ 3) * B(k+ 3,j) ;
        temp += A(i,k+ 4) * B(k+ 4,j) ;
        temp += A(i,k+ 5) * B(k+ 5,j) ;
        temp += A(i,k+ 6) * B(k+ 6,j) ;
        temp += A(i,k+ 7) * B(k+ 7,j) ;
        temp += A(i,k+ 8) * B(k+ 8,j) ;
        temp += A(i,k+ 9) * B(k+ 9,j) ;
        temp += A(i,k+10) * B(k+10,j) ;
        temp += A(i,k+11) * B(k+11,j) ;
        temp += A(i,k+12) * B(k+12,j) ;
        temp += A(i,k+13) * B(k+13,j) ;
        temp += A(i,k+14) * B(k+14,j) ;
        temp += A(i,k+15) * B(k+15,j) ;
      }
	    for ( ; k < p ; ++k ) temp += A(i,k)*B(k,j) ;
	    C(i,j) = temp ;
    }
  }
}

///////////////////////////////////////////////////////

void
MM_pointer( valueType const a[], indexType lda, // n x p
            valueType const b[], indexType ldb, // p x m
            valueType       c[], indexType ldc,
            indexType n,
            indexType p,
            indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i ) {
    valueType * pc = &C(i,0) ;
    for ( indexType j = 0 ; j < m ; ++j ) {
	    valueType temp = 0 ;
      valueType const * pa = &A(i,0) ;
      valueType const * pb = &B(0,j) ;
	    for ( indexType k = 0 ; k < p ; ++k ) {
        temp += (*pa) * (*pb) ;
        NEXT_COLUMN(a) ;
        NEXT_ROW(b) ;
      }
      *pc = temp ;
	    NEXT_COLUMN(c) ;
    }
  }
}

/* from Monica Lam ASPLOS-IV paper */
void
MM_lam( valueType const a[], indexType lda, // n x p
        valueType const b[], indexType ldb, // p x m
        valueType       c[], indexType ldc,
        indexType n,
        indexType p,
        indexType m ) {

  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType j = 0 ; j < m ; ++j )
      C(i,j) = 0 ;

  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType k = 0 ; k < p ; ++k ) {
      valueType a_entry = A(i,k) ;
      for ( indexType j = 0 ; j < m ; ++j )
        C(i,j) += a_entry*B(k,j) ;
    }
}

/* from Monica Lam ASPLOS-IV paper */
void
MM_tiling( valueType const a[], indexType lda, // n x p
           valueType const b[], indexType ldb, // p x m
           valueType       c[], indexType ldc,
           indexType n,
           indexType p,
           indexType m,
           indexType step ) {

  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType j = 0 ; j < m ; ++j )
      C(i,j) = 0 ;

  for( indexType kk = 0 ; kk < p ; kk += step )
    for( indexType jj = 0 ; jj < m ; jj += step )
      for ( indexType i = 0 ; i < n ; ++i )
        for ( indexType k = kk ; k < min(kk+step,p) ; ++k ) {
          valueType a_entry = A(i,k) ;
          for ( indexType j = jj ; j < min(jj+step,m) ; ++j )
            C(i,j) += a_entry*B(k,j) ;
        }
}

/* Matrix Multiply tuned for SS-10/30;
 *                      Maeno Toshinori
 *                      Tokyo Institute of Technology
 *
 * Using gcc-2.4.1 (-O2), this program ends in 12 seconds on SS-10/30. 
 *
 * in original algorithm - sub-area for cache tiling
 * #define      L       20
 * #define      L2      20
 * three 20x20 matrices reside in cache; two may be enough
 */
void
MM_maeno( valueType const a[], indexType lda, // n x p
          valueType const b[], indexType ldb, // p x m
          valueType       c[], indexType ldc,
          indexType n,
          indexType p,
          indexType m,
          indexType lparm ) {

  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType j = 0 ; j < m ; ++j )
      C(i,j) = 0 ;

	for ( indexType i2 = 0 ; i2 < n ; i2 += lparm )
	  for ( indexType kk = 0 ; kk < p ; kk += lparm ) {
	    indexType it = i2 + lparm ;
	    indexType kt = kk + lparm ;
      for ( indexType j = 0 ; j < m ; j += 4 )
        for( indexType i = i2 ; i < min(n,it) ; i += 2 ) {
	        valueType t0=0,
                    t1=0,
                    t2=0,
                    t3=0,
                    t4=0,
                    t5=0,
                    t6=0,
                    t7=0 ;
          for( indexType k = kk ; k < min(p,kt) ; ++k  ) {
            valueType s = A(i,k) ;
            t0 += s*B(k,j+0);
            t1 += s*B(k,j+1);
            t2 += s*B(k,j+2);
            t3 += s*B(k,j+3);
            s = A(i+1,k) ;
            t4 += s*B(k,j+0);
            t5 += s*B(k,j+1);
            t6 += s*B(k,j+2);
            t7 += s*B(k,j+3);
          }
          C(i+0,j+0) += t0 ;
          C(i+0,j+1) += t1 ;
          C(i+0,j+2) += t2 ;
          C(i+0,j+3) += t3 ;
          C(i+1,j+0) += t4 ;
          C(i+1,j+1) += t5 ;
          C(i+1,j+2) += t6 ;
          C(i+1,j+3) += t7 ;
		    }
	  }
}

/*
 * Matrix Multiply by Dan Warner, Dept. of Mathematics, Clemson University
 *
 *    mmbu2.f multiplies matrices a and b
 *    a and b are n by n matrices
 *    nb is the blocking parameter.
 *    the tuning guide indicates nb = 50 is reasonable for the
 *    ibm model 530 hence 25 should be reasonable for the 320 
 *    since the 320 has 32k rather than 64k of cache.      
 *    Inner loops unrolled to depth of 2
 *    The loop functions without clean up code at the end only
 *    if the unrolling occurs to a depth k which divides into n
 *    in this case n must be divisible by 2.
 *    The blocking parameter nb must divide into n if the
 *    multiply is to succeed without clean up code at the end. 
 *
 * converted to c by Mark Smotherman
 * note that nb must also be divisible by 2 => cannot use 25, so use 20
 */

void
MM_warner( valueType const a[], indexType lda, // n x p
           valueType const b[], indexType ldb, // p x m
           valueType       c[], indexType ldc,
           indexType n,
           indexType p,
           indexType m,
           indexType nb ) {

	for( indexType ii = 0 ; ii < n ; ii += nb ) {
	  for( indexType jj = 0 ; jj < m ; jj += nb ) {

	    for( indexType i = ii ; i < min(n,ii + nb) ; ++i )
	      for( indexType j = jj ; j < min(m,jj + nb) ; ++j )
          C(i,j) = 0.0;

      for( indexType kk = 0 ; kk < p ; kk += nb ) {
        for( indexType i = ii ; i < min(n-1,ii + nb) ; i += 2 ) {
          for( indexType j = jj ; j < min(m-1,jj + nb) ; j += 2 ) {
            valueType s00 = C(i+0,j+0) ;
            valueType s01 = C(i+0,j+1) ;
            valueType s10 = C(i+1,j+0) ;
            valueType s11 = C(i+1,j+1) ;
            for( indexType k = kk ; k < kk + nb ; ++k ){
              s00 += A(i+0,k)*B(k,j+0) ;
              s01 += A(i+0,k)*B(k,j+1) ;
              s10 += A(i+1,k)*B(k,j+0) ;
              s11 += A(i+1,k)*B(k,j+1) ;
            }
            C(i+0,j+0) = s00 ;
            C(i+0,j+1) = s01 ;
            C(i+1,j+0) = s10 ;
            C(i+1,j+1) = s11 ;
          }
	      }
	    }
	  }
	}
}

///////////////////////////////////////////////////////

void
MM_blas( valueType const a[], indexType lda, // n x p
         valueType const b[], indexType ldb, // p x m
         valueType       c[], indexType ldc,
         indexType n,
         indexType p,
         indexType m ) {
#ifdef ROW_MAJOR
  alglin::gemm( "N", "N", n, m, p, 
                 1.0, b, ldb,
                      a, lda, 
                 0.0, c, ldc ) ;
#else
  alglin::gemm( "N", "N", n, m, p, 
                 1.0, a, lda,
                      b, ldb, 
                 0.0, c, ldc ) ;
#endif
}

///////////////////////////////////////////////////////

void
MM_blocking( valueType const a[], indexType lda, // n x p
             valueType const b[], indexType ldb, // p x m
             valueType       c[], indexType ldc,
             indexType n,
             indexType p,
             indexType m,
             indexType step ) {

  for ( indexType i = 0 ; i < n ; ++i )
    for ( indexType j = 0 ; j < m ; ++j )
      C(i,j) = 0 ;

  for( indexType jj = 0 ; jj < m ; jj += step )
    for( indexType ii = 0 ; ii < n ; ii += step )
      for( indexType kk = 0 ; kk < p ; kk += step )
        for( indexType j = jj ; j < min(m,jj+step) ; ++j )
          for( indexType i = ii ; i < min(n,ii+step) ; ++i ) {
            valueType temp = C(i,j) ;
            for( indexType k = kk ; k < min(p,kk+step) ; ++k )
              temp += A(i,k)*B(k,j) ;
            C(i,j) = temp ;
          }
}

///////////////////////////////////////////////////////

#include <iostream>

static valueType * aux = NULL ;
static indexType naux  = 5000*5000*10 ;

void
MM_strassen( valueType const A[], indexType lda, // n x p
             valueType const B[], indexType ldb, // p x m
             valueType       C[], indexType ldc,
             indexType n,
             indexType p,
             indexType m ) {

  if ( aux == NULL ) aux = new valueType [ naux ] ;

#ifdef ROW_MAJOR
  strassen::matrixMultiply( n,       /* Rows of C and A                       */
                            m,       /* Cols of B and C                       */
                            p,       /* Cols of A and Rows of B               */
                            B,       /* Pointer to coefficients of A          */
                            ldb,     /* Leading dimension of A, Fortran style */
                            A,       /* Pointer to coefficients of B          */
                            lda,     /* Leading dimension of B, Fortran style */
                            C,       /* Pointer to coefficients of C          */
                            ldc,     /* Leading dimension of C, Fortran style */
                            aux,     /* Pointer to auxillary storage or NULL  */
                            naux ) ; /* Number of elements in aux or <= 0     */
#else
  strassen::matrixMultiply( n,       /* Rows of C and A                       */
                            m,       /* Cols of B and C                       */
                            p,       /* Cols of A and Rows of B               */
                            A,       /* Pointer to coefficients of A          */
                            lda,     /* Leading dimension of A, Fortran style */
                            B,       /* Pointer to coefficients of B          */
                            ldb,     /* Leading dimension of B, Fortran style */
                            C,       /* Pointer to coefficients of C          */
                            ldc,     /* Leading dimension of C, Fortran style */
                            aux,     /* Pointer to auxillary storage or NULL  */
                            naux ) ; /* Number of elements in aux or <= 0     */
#endif
}

// EOF mm.hh
