#ifndef STRASSEN_HH
#define STRASSEN_HH

#include <sstream>

#ifndef ASSERT
  #define ASSERT(COND,MSG) \
    if ( !(COND) ) { std::ostringstream ost ; ost << MSG << '\n' ; throw std::runtime_error(ost.str()) ; }
#endif

#ifndef REAL_TYPE
#define REAL_TYPE double
#endif

#ifndef INTEGER_TYPE
#define INTEGER_TYPE int
#endif

namespace strassen {

  typedef REAL_TYPE    valueType ;
  typedef INTEGER_TYPE indexType ;

  void
  matrixMultiply( indexType const m,     /* Rows of C and A                       */
                  indexType const n,     /* Cols of B and C                       */
                  indexType const k,     /* Cols of A and Rows of B               */
                  valueType const A[],   /* Pointer to coefficients of A          */
                  indexType const lda,   /* Leading dimension of A, Fortran style */
                  valueType const B[],   /* Pointer to coefficients of B          */
                  indexType const ldb,   /* Leading dimension of B, Fortran style */
                  valueType       C[],   /* Pointer to coefficients of C          */
                  indexType const ldc,   /* Leading dimension of C, Fortran style */
                  valueType       aux[], /* Pointer to auxillary storage or NULL  */
                  indexType       naux   /* Number of elements in aux or <= 0     */ ) ;

}

#endif
